/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "pic32_c.h"
#include "init.h"
#include "debug.h"

/* Specific functionality. */

#include "devconfig.h"
#include "display.h"
#include "display_config.h"
#include "font.h"
#include "main.h"
#include "vga.h"
#include "vga_display.h"



/* Define timers if not indicated in the build configuration. */

#ifndef LINE_CHANNELS
#define LINE_CHANNELS   1
#endif

#ifndef LINE_TIMER
#define LINE_TIMER      2
#endif

#ifndef TRANSFER_TIMER
#define TRANSFER_TIMER  0
#endif

/* Define different output ports and pins for parallel mode. */

#ifndef PARALLEL_MODE
#define VGA_OUTPUT      PORTB
#define OC1_PIN         RPA0R
#define OC2_PIN         RPA1R
#define LED_PIN         (1 << 3)
#else
#define VGA_OUTPUT      PM_REG(0, PMxDIN)
#define OC1_PIN         RPB4R
#define OC2_PIN         RPB5R
#define LED_PIN         (1 << 2)
#endif



/* Bundled image and font data. */

extern uint8_t screendata[];
extern uint32_t screendata_width, screendata_height;
extern uint8_t sprite[];
extern uint32_t sprite_width, sprite_height;

extern uint8_t fontchars[];
extern uint32_t fonttable[];
extern uint32_t fontbase, fontlimit;

static font_config_t font_config;



/* Busy wait. */

static void wait(uint32_t delay)
{
    uint32_t counter = delay;

    if (!delay) return;
    while (counter--) __asm__("");	/* retain loop */
}

/* Blink an attached LED with delays implemented using a loop. */

static void blink(uint32_t delay, uint32_t port, uint32_t pins)
{
    /* Clear outputs (LED). */

    CLR_REG(port, pins);

    while (1)
    {
        wait(delay);

        /* Invert outputs (LED). */

        INV_REG(port, pins);
    }
}

/* Move a sprite around on the framebuffer. */

static void animate(uint32_t delay)
{
    uint8_t background[sprite_width * sprite_height];
    int x, y;
    int dir[] = {1, 0, -1, 0, 1}, i = 0, width, column_width;
    int xsource, xdisplay, xorigin = 0, yorigin = 0;

    while (1)
    {
        for (y = 0; y < screendata_height - sprite_height; y++)
        {
            for (x = 0; x < screendata_width - sprite_width; x++)
            {
                /* Copy to the store from the display, then blit the image. */

                copy_display(&display_config, background,
                             sprite_width, sprite_height,
                             x, y, -1, 0);
                copy_display(&display_config, sprite,
                             sprite_width, sprite_height,
                             x, y, 0x8c, 1);

                wait(delay);

                /* Copy to the display from the store, restoring the original
                   background. */

                copy_display(&display_config, background,
                             sprite_width, sprite_height,
                             x, y, -1, 1);

                /* Scroll in the indicated direction. */

                scroll_display(&display_config, dir[i], dir[i + 1]);

                /* Update the vertical origin if appropriate. */

                if (dir[i + 1])
                {
                    yorigin += dir[i + 1];

                    if (yorigin < 0)
                        yorigin += screendata_height;
                    else if (yorigin >= screendata_height)
                        yorigin -= screendata_height;
                }

                /* For horizontal scrolling, plot the exposed column at the left
                   (if scrolling left) or at the right (if scrolling right). */

                if (dir[i])
                {
                    /* Due to the effect of a simple screen start increment in
                       the dual channel configuration, horizontal scrolling
                       involves two pixel increments and thus requires a two-
                       pixel column to be plotted. */

                    width = dir[i] * LINE_CHANNELS;
                    column_width = width < 0 ? -width : width;

                    /* Plot either at the left or right edge. */

                    xdisplay = width < 0 ? 0 : screendata_width - width;

                    /* Determine the location of the column to be plotted. */

                    xorigin += width;

                    if (xorigin < 0)
                        xorigin += screendata_width;
                    else if (xorigin >= screendata_width)
                        xorigin -= screendata_width;

                    xsource = (xdisplay + xorigin) % screendata_width;

                    /* Plot a column in two pieces if the vertical origin is
                       non-zero. The first piece is at (xdisplay, 0) and
                       provides the lower part of the background image displaced
                       upwards (or downwards having wrapped around) on the
                       screen. */

                    copy_display_section(&display_config, screendata,
                                         screendata_width, screendata_height,
                                         xsource, yorigin,
                                         column_width, screendata_height - yorigin,
                                         xdisplay, 0,
                                         -1, 1);

                    /* The second column is at (xdisplay, h - yorigin) and
                       provides the upper part of the background image displaced
                       downwards (or upwards having wrapped around) on the
                       screen. */

                    if (yorigin)
                        copy_display_section(&display_config, screendata,
                                             screendata_width, screendata_height,
                                             xsource, 0,
                                             column_width, yorigin,
                                             xdisplay, screendata_height - yorigin,
                                             -1, 1);
                }
            }

            /* Switch direction periodically. */

            i++;
            if (i == 4)
                i = 0;
        }
    }
}

/* Fill the screen with characters. */

static void write_chars(void)
{
    int x = 0, y = 0;
    char c;

    font_config.chars = (char_definition_t *) fontchars;
    font_config.table = fonttable;
    font_config.base = fontbase;
    font_config.limit = fontlimit;

    while (y < display_config.line_count)
        for (c = (char) font_config.base; c < (char) font_config.limit; c++)
        {
            x = write_char(&display_config, &font_config, c, x, y, 0xff);

            if (x > display_config.line_length)
            {
                x = 0; y += 9;
            }
        }
}



/* Main program. */

void main(void)
{
    init_memory();
    init_pins();
    init_outputs();

    unlock_config();
    config_oc();
    config_uart();
    lock_config();

    init_dma();

#ifdef PARALLEL_MODE
    init_pm();

    /* Configure parallel master mode. */

    pm_init(0, 0b10);
    pm_set_output(0, 1, 0);
    pm_on(0);
#endif

    /* Initialise VGA output with one or two line channels, configuring a line
       timer and any transfer timer, with an initiating channel being introduced
       if a transfer timer is specified. */

    init_vga_with_timers(&display_config, LINE_CHANNELS, LINE_TIMER, TRANSFER_TIMER);

    /* Configure VGA output transfer to the output register, also configuring
       output compare units for horizontal and vertical sync. */

    vga_configure_transfer(VGA_OUTPUT);
    vga_configure_sync(1, 2);

    uart_init(1, FPB, 115200);
    uart_on(1);

    interrupts_on();

    /* Plot the image centred on the screen. */

    copy_display(&display_config, screendata, screendata_width, screendata_height,
                 (LINE_LENGTH - screendata_width) / 2,
                 (LINE_COUNT - screendata_height) / 2, -1, 1);

    /* Write a sequence of characters. */

    write_chars();

    /* Move a sprite around on the screen with a delay between each movement. */

    animate(1 << 21);
}



/* Exception and interrupt handlers. */

void exception_handler(void)
{
    blink(1 << 22, PORTA, LED_PIN);
}

void interrupt_handler(void)
{
    uint32_t ifs;

    /* Check for a OC1 interrupt condition. */

    ifs = REG(OCIFS) & OC_INT_FLAGS(1, OCxIF);

    if (ifs)
    {
        vga_interrupt_handler();
        CLR_REG(OCIFS, ifs);
    }
}



/* Peripheral pin configuration. */

void config_oc(void)
{
    /* Map OC1. */

    REG(OC1_PIN) = 0b0101;

    /* Map OC2. */

    REG(OC2_PIN) = 0b0101;
}

void config_uart(void)
{
    /* Map U1RX to RPB13. */

    REG(U1RXR) = 0b0011;        /* U1RXR<3:0> = 0011 (RPB13) */

    /* Map U1TX to RPB15. */

    REG(RPB15R) = 0b0001;       /* RPB15R<3:0> = 0001 (U1TX) */

    /* Set RPB13 to input. */

    SET_REG(TRISB, 1 << 13);
}
